import os, pygame, random, math
import ai
from pygame.locals import *

def max(a,b):
	if a>b: return a
	else: return b

def clamp(a,mn,mx):
	if a>mx: return mx
	if a<mn: return mn
	return a

# Cell Types	
CellBlank = 0
CellWhite = 1
CellBlack = 2

# Player Types
PlayerHuman = 0
PlayerComputer = 1

def OppositePlayer( Type ):
	if Type == CellWhite: return CellBlack
	if Type == CellBlack: return CellWhite
	return CellBlank

class CellType:
	"""A single cell in an othello board."""

	def __init__(self):
		self.Type = CellBlank
		self.AnimTo = 0      # Cell type being transitioned to (CellBlank for not animating).
		self.StartFrame = 0  # Starting animation frame in board anim frame units.  Future values are okay.

	def Copy(self):
		newCell = CellType()
		newCell.Type = self.Type
		newCell.AnimTo = self.AnimTo
		newCell.StartFrame = self.StartFrame
		return newCell

	def Animate(self, To, Start):
		self.Type = To       # Type is set immediately.
		self.AnimTo = To
		self.StartFrame = Start

class MoveType:
	"""Represents a valid move.  These are generated by a GameStateType object."""

	def __init__(self):
		self.X = 0
		self.Y = 0

	def __init__(self, _X, _Y):
		self.X = _X
		self.Y = _Y

	def SetInvalid():
		self.X = -1

	def IsValid():
		return self.X != -1

class GameStateType:
	"""Represents the state of the game."""

	def __init__(self):
		self.Width = 8
		self.Height = 8
		self.Cells = []

		self.MaxMoves = 0
		self.MoveNumber = 0

		self.CurPlayer = CellWhite

		self.BlackScore = 0
		self.WhiteScore = 0

		self.AnimFrame = 0

		self.NewGame()

	def Copy(self):
		newState = GameStateType()
		newState.Width = self.Width
		newState.Height = self.Height
		newState.Cells = []
		for row in self.Cells:
			newRow = []
			for cell in row:
				newCell = cell.Copy()
				newRow.append(newCell)
			newState.Cells.append(newRow)

		newState.MaxMoves = self.MaxMoves
		newState.MoveNumber = self.MoveNumber

		newState.CurPlayer = self.CurPlayer

		newState.BlackScore = self.BlackScore
		newState.WhiteScore = self.WhiteScore

		newState.AnimFrame = self.AnimFrame
		
		return newState
		
	def InBounds(self, X, Y):
		return X >= 0 and X < self.Width and Y >= 0 and Y < self.Height

	def AttemptMove(self, X, Y, TestOnly):
		# Cannot click on a cell that is already taken.
		if self.Cells[X][Y].Type != CellBlank:
			return False

		Valid = False

		# Check for runs in every possible direction.
		for dx in range(-1, 2):
			for dy in range(-1, 2):

				if dx == 0 and dy == 0: 
					continue

				# See how far the run goes.
				len = 0
				cx = X + dx
				cy = Y + dy
				while self.InBounds(cx, cy) and self.Cells[cx][cy].Type == OppositePlayer(self.CurPlayer):
					cx += dx
					cy += dy
					len += 1

				# Run must cover at least one square, and end with the current player type.
				if not self.InBounds(cx, cy):
					continue
				if len == 0:
					continue
				if self.Cells[cx][cy].Type != self.CurPlayer:
					continue

				# We now have a valid run.  Either return success, or else apply the move.
				if TestOnly:
					return True

				# Apply the move, converting the pieces in the run.  Note that this happens for each direction included in the run.
				len = 0
				cx = X + dx
				cy = Y + dy
				while self.InBounds(cx, cy) and self.Cells[cx][cy].Type == OppositePlayer(self.CurPlayer):
					self.Cells[cx][cy].Animate( self.CurPlayer, self.AnimFrame + 10 * len );
					cx += dx
					cy += dy
					len += 1

				Valid = True

		if Valid:
			self.Cells[X][Y].Type = self.CurPlayer
			self.CurPlayer = OppositePlayer( self.CurPlayer )
			self.MoveNumber += 1
			return True
		else:
			return False

	def CheckMove( self, Move ):
		return self.AttemptMove( Move.X, Move.Y, True )

	def ApplyMove( self, Move ):
		self.AttemptMove( Move.X, Move.Y, False )

	def NewGame( self ):
		self.Cells = []
		for x in range(0, self.Width):
			self.Cells.append([])
			for y in range(0, self.Height):
				Cell = CellType()
				Cell.AnimTo = CellBlank
				Cell.Type = CellBlank
				self.Cells[x].append( Cell )
 
		cx = self.Width/2
		cy = self.Height/2
		self.Cells[cx-1][cy].Type = CellWhite
		self.Cells[cx-1][cy-1].Type = CellBlack
		self.Cells[cx][cy].Type = CellBlack
		self.Cells[cx][cy-1].Type = CellWhite
		self.MoveNumber = 0
		self.CurPlayer = CellWhite

	def Evaluate( self, Player ):
		self.BlackScore = 0
		self.WhiteScore = 0
		for x in range(0, self.Width):
			for y in range(0, self.Height):
				if self.Cells[x][y].Type == CellWhite: self.WhiteScore += 1
				if self.Cells[x][y].Type == CellBlack: self.BlackScore += 1
	
		if Player == CellWhite:
			Value = self.WhiteScore - self.BlackScore
		else:
			Value = self.BlackScore - self.WhiteScore

		# Factor corner pieces.
		if self.Cells[0][0].Type == Player: 
			Value += 4 
		else: 
			Value -= 4
		if self.Cells[self.Width-1][0].Type == Player: 
			Value += 4 
		else: 
			Value -= 4
		if self.Cells[0][self.Height-1].Type == Player: 
			Value += 4 
		else: 
			Value -= 4
		if self.Cells[self.Width-1][self.Height-1].Type == Player: 
			Value += 4 
		else: 
			Value -= 4

		# Factor edge pieces.  Relies on Width==Height.
		for x in range(1, self.Width-1):
			if self.Cells[x][0].Type == Player: 
				Value += 2 
			else: 
				Value -= 2
			if self.Cells[x][self.Height-1].Type == Player: 
				Value += 2 
			else: 
				Value -= 2
			if self.Cells[0][x].Type == Player: 
				Value += 2 
			else: 
				Value -= 2
			if self.Cells[self.Width-1][x].Type == Player: 
				Value += 2 
			else: 
				Value -= 2

		return Value

	def GenerateMoves( self ):
		Moves = []
		for x in range(0, self.Width):
			for y in range(0, self.Height):
				if self.AttemptMove( x, y, True ):
					Moves.append( MoveType( x, y ) )
		return Moves

	def IsFinished( self ):
		Moves = self.GenerateMoves()
		return len(Moves) == 0
	
	def IsAnimating( self ):
		for i in range(0, self.Width):
			for j in range(0, self.Height):
				if self.Cells[i][j].AnimTo != CellBlank:
					return True
		return False

	def IsMyTurn( self, Player ):
		return Player == self.CurPlayer

	def GetTurnCount( self ):
		return self.MoveNumber

class GameType:
	def __init__( self ):
		self.State = GameStateType()
		self.PlayerType = [ 0, PlayerComputer, PlayerComputer ]
		self.screen = None
		self.Bitmaps = []
		for x in range(0, 26):
			self.Bitmaps.append( pygame.image.load( "images/flip%04d.bmp" % x ) )
		self.Rect = pygame.Rect(200, 100, 800, 600)
		
	def Paint( self ):
		pygame.draw.rect( self.screen, ( 255, 255, 255 ), self.Rect )

		xstep = self.Rect.width / self.State.Width
		ystep = self.Rect.height / self.State.Height

		Drawn = False
		for i in range(0, self.State.Width):
			for j in range(0, self.State.Height):
				Cell = self.State.Cells[i][j]
				
				Frame = self.State.AnimFrame - Cell.StartFrame

				Bitmap = 0
				if Cell.AnimTo == CellBlack:
					Bitmap = max(12, 12+Frame)
				elif Cell.AnimTo == CellWhite:
					Bitmap = max(0, Frame)
				else:
					if Cell.Type == CellBlack:
						Bitmap = 0
					elif Cell.Type == CellWhite:
						Bitmap = 12

				if Cell.Type != CellBlank:
					self.screen.blit( self.Bitmaps[Bitmap%25], ( self.Rect.left + (i+0.5) * xstep - 16, self.Rect.top + (j+0.5) * ystep - 16 ) )

				# Stop animation here.
				if Frame > 12:
					Cell.AnimTo = CellBlank

		for i in range(0, self.State.Width+1):
			pygame.draw.line( self.screen, ( 0, 0, 0 ), (self.Rect.left + i*xstep, self.Rect.top + 0), (self.Rect.left + i*xstep, self.Rect.top + self.State.Height*ystep) )

		for i in range(0, self.State.Height+1):
			pygame.draw.line( self.screen, ( 0, 0, 0 ), (self.Rect.left + 0, self.Rect.top + i*ystep), (self.Rect.left + self.State.Width*xstep, self.Rect.top + i*ystep) )

	def OnClick( self, pos ):
		if self.State.IsFinished():
			return
		if self.PlayerType[self.State.CurPlayer] != PlayerHuman:
			return
		if self.State.IsAnimating():
			return
			
		x = pos[0] * self.State.Width / self.screen.get_width()
		y = pos[1] * self.State.Height / self.screen.get_height()

		self.State.AttemptMove( x, y, False )

	def OnAnimTimer( self ):
		self.State.AnimFrame += 1
		if not self.State.IsAnimating():
			if self.State.IsFinished():
				self.Status = "Game Over"
				#self.FinishGame()
			else:
				if self.State.CurPlayer == CellWhite:
					self.Status = "White's Move"
				else:
					self.Status = "Black's Move"

	def OnAITimer( self ):
		if self.State.IsFinished() or self.State.IsAnimating(): 
			return
	    
		if self.PlayerType[self.State.CurPlayer] == PlayerComputer:
			Move = ai.GetMove( self.State, self.State.CurPlayer, 4 )
			#Move = ai.GetRandomMove( self.State, self.State.CurPlayer )
			#if Move is not None:
			self.State.AttemptMove( Move.X, Move.Y, False )

		#if PlayerType[self.State.CurPlayer] == PlayerComputer:
		#	AI.GetRandomMove( Game, Game.CurPlayer, Move );
		#else:
		#	if PlayerType[Game.CurPlayer] == plComputerMedium:
		#		AI.CutoffDepth = 2
		#	elif PlayerType[Game.CurPlayer] == plComputerHard:
		#		AI.CutoffDepth = 6
		#Move = AI.GetMove( Game, Game.CurPlayer )
		#self.State.AttemptMove( Move.X, Move.Y, False )

Game = GameType()

def	main():
	pygame.init()
	
	Game.screen = pygame.display.set_mode((1200, 825))

	Game.font = pygame.font.Font(None, 36)

	pygame.display.set_caption('GameJam test')

	clock =	pygame.time.Clock()

	while True:
		clock.tick(30)

		for	event in pygame.event.get():
			if event.type == QUIT:
				return
			elif event.type	== KEYDOWN and event.key ==	K_ESCAPE:
				return
			elif event.type == MOUSEBUTTONDOWN:
				Game.MouseDown = 1
			elif event.type == MOUSEBUTTONUP:
				Game.OnClick( event.pos )
				Game.MouseDown = 0
			elif event.type == MOUSEMOTION:
				Game.MouseX = event.pos[0]
				Game.MouseY = event.pos[1]

		pygame.draw.rect(Game.screen, (64, 64, 64), pygame.Rect(0, 0, Game.screen.get_width(), Game.screen.get_height()))

		Game.Paint()		
		pygame.display.flip()
		
		Game.OnAnimTimer()
		Game.OnAITimer()
		if Game.State.IsFinished():
			Game.State.NewGame()

if __name__	== '__main__': main()
